""" Inverse optimization for Linear Models

This module contains a collection of linear inverse optimization models. It
includes the ones covered by generalized inverse optimization, including the
absolute and relative duality gap, and p-norm methods. As we add new models,
we should include them in the __main__ file.
"""



from .RelativeDualityGap import RelativeDualityGap
from .AbsoluteDualityGap import AbsoluteDualityGap
from .pNorm import pNorm



""" Decision Space Inverse Optimization with a p-norm.

These models measure error in the space of decision variables, rather than
objective values. In particular, these models aim to identify a cost vector that
induces optimal decisions for the forward problem that are of minimum aggregate
distance to the corresponding observed decisions. Any norm can be used in the
decision space, but the key distinction is that the imputed optimal decisions
must be feasible for the forward problem. A decision space inverse optimization
problem is formulated as

.. math::

    \min_{\mathbf{c, y}, \\boldsymbol{ \epsilon_1,} \dots, \\boldsymbol{ \epsilon_Q }} \quad  & \sum_{q=1}^Q \| \\boldsymbol{\epsilon_q} \|

    \\text{s.t.}\quad\quad  & \mathbf{A'y = c}

    & \mathbf{c'\hat{x}_q = b'y} + \\boldsymbol{\epsilon_q}, \quad \\forall q

    & \mathbf{A ( \hat{x}_q -} \\boldsymbol{ \epsilon_q } \mathbf{ ) \geq b}

    & \| \mathbf{c} \|_1 = 1

    & \mathbf{y \geq 0}
"""
import cvxpy as cvx
import numpy as np
#import pudb

from ..utils.invoutils import validateFOP


class pNorm():
    """ Formulate a Decision Space (p-norm) method of GMIO.

    Args:
        tol (int): Sets number of significant digits. Default is 8.
        p (int): Sets p for lp norm. Can be integer or 'inf'. Default is 2.
        verbose (bool): Sets displays.  Default is False.
        ban_constraints (list): A list of constraint indices to force to zero when solving. Default is none.

    Example:
        Suppose that the variables ``A`` and ``b`` are numpy matrices and ``points`` is
        a list of numpy arrays::

           model = pNorm(p=2)
           model.FOP(A, b)
           model.solve(points)
           print (model.c)
    """

    def __init__(self, **kwargs):
        """
        """
        self._fop = False
        self._verbose = False
        self._solved = False
        self.p = 2
        self.tol = 8
        self.ban_constraints = []
        self._kwargs = self._initialize_kwargs(kwargs)

    def FOP(self, A, b):
        """ Create a forward optimization problem.

        Args:
            A (matrix): numpy matrix of shape :math:`m \\times n`.
            b (matrix): numpy matrix of shape :math:`m \\times 1`.

        Currently, the forward problem is constructed by the user supplying a
        constraint matrix `A` and vector `b`. The forward problem is

        .. math::

            \min_{\mathbf{x}} \quad&\mathbf{c'x}

            \\text{s.t} \quad&\mathbf{A x \geq b}
        """
        #self.A = np.mat(A)
        #self.b = np.mat(b)
        self.A, self.b = validateFOP(A, b)
        self._fop = True

    def solve(self, points, **kwargs):
        """ Solves the inverse optimization problem.

        Args:
            points (list): list of numpy arrays, denoting the (optimal) observed points.

        Returns:
            error (float): the optimal value of the inverse optimization problem.

        To solve a decision space inverse optimization problem, we solve the
        following convex problem for every constraint

        .. math::

            \min_{\\boldsymbol{ \epsilon_1,} \dots, \\boldsymbol{ \epsilon_Q }} \quad  & \sum_{q=1}^Q \| \\boldsymbol{\epsilon_q} \|_p

            \\text{s.t.} \quad\quad  & \mathbf{ a_i'(x_q - }\\boldsymbol{\epsilon_q}\mathbf{)} = b_i

            & \mathbf{ A ( x_q - }\\boldsymbol{\epsilon_q}\mathbf{) \geq b}
        """
        self._kwargs = self._initialize_kwargs(kwargs)

        points = [np.mat(point).T for point in points]
        assert self._fop, 'No forward model given.'
        self.error = self._solveFeasibleProjection(points)
        return self.error

    def _solveFeasibleProjection(self, points):
        m, n = self.A.shape
        bestResult = np.inf

        for i in range(m):
            if i in self.ban_constraints:
                result = np.inf
            else:
                ai = self.A[i]
                bi = self.b[i]
                result = self._project_to_hyperplane(points, ai, bi)

            if result < bestResult:
                bestResult = result
                self.dual = np.zeros(m)
                self.dual[i] = 1.0 / np.linalg.norm(ai, np.inf)
                self.c = ai / np.linalg.norm(ai, np.inf)
        self._solved = True
        #self.dual = self.dual.T.tolist()[0]
        self.c = self.c.tolist()[0]
        self.error = bestResult
        return result

    def _project_to_hyperplane(self, points, ai, bi):
        """ Helper function that solves feasible projection to a hyperplane
        """
        m, n = self.A.shape
        nPoints = len(points)
        epsilons = [cvx.Variable(n) for pt in points]
        objFunc = []
        cons = []
        for x in range(nPoints):
            objFunc.append(cvx.norm(epsilons[x], self.p))
            cons.append(self.A * (points[x] - epsilons[x]) >= self.b)
            cons.append(ai * (points[x] - epsilons[x]) == bi)
        obj = cvx.Minimize(sum(objFunc))
        prob = cvx.Problem(obj, cons)
        result = prob.solve()
        return result

    def optimal_points(self, points):
        """ Get the projected optimal points.

        Args:
            points (list): list of numpy arrays, denoting the observed points.

        Returns:
            truePoints (list): list of numpy arrays denoting the imputed optimal points.

        Once an inverse optimization problem is solved and the forward model is
        completed, you can take a collection of observed data points and get
        the 'imputed' optimal points.
        """
        if self._solved:
            m, n = self.A.shape
            ind = np.argmax(self.dual)
            ai = self.A[ind]
            bi = self.b[ind]
            epsilons = [cvx.Variable(n) for pt in points]
            objFunc = []
            cons = []
            for x, point in enumerate(points):
                objFunc.append(cvx.norm(epsilons[x], self.p))
                cons.append(self.A * (points[x] - epsilons[x]) >= self.b)
                cons.append(ai * (points[x] - epsilons[x]) == bi)
            obj = cvx.Minimize(sum(objFunc))
            prob = cvx.Problem(obj, cons)
            result = prob.solve()
            xStars = [(np.mat(point).T - epsilon.value)
                      for point, epsilon in zip(points, epsilons)]
            xStars = [xs.T.tolist()[0] for xs in xStars]
        else:
            self.solve(points)
            xStars = self.optimal_points(points)
        return xStars

    def rho(self, points):
        """ Solves the goodness of fit.
        """
        assert self._solved, 'you need to solve first.'

        m, n = self.A.shape
        projections = self.optimal_points(points)
        _pts = [np.mat(pt).T for pt in points]
        numer = [
            np.linalg.norm(pj - pt, self.p)
            for pj, pt in zip(projections, _pts)
        ]
        numer = sum(numer)
        denom = 0
        for i in range(m):
            ai = self.A[i]
            bi = self.b[i]
            result = self._project_to_hyperplane(points, ai, bi)
            denom += result
        rho = 1 - numer / denom
        return rho

    def _initialize_kwargs(self, kwargs):
        # common kwargs
        if 'verbose' in kwargs:
            assert isinstance(kwargs['verbose'],
                              bool), 'verbose needs to be True or False.'
            self._verbose = kwargs['verbose']
        if 'tol' in kwargs:
            assert isinstance(kwargs['tol'],
                              int), 'tolerance needs to be an integer.'
            self.tol = kwargs['tol']

        # class specific kwargs
        if 'p' in kwargs:
            assert isinstance(
                kwargs['p'],
                int) or kwargs['p'] is 'inf', 'p needs to be an integer'
            self.p = kwargs['p']

        return kwargs



""" Absolute Duality Gap Inverse Optimization

The absolute duality gap method for inverse optimization minimizes the aggregate
duality gap between the primal and dual objective values for each observed
decision. The problem is formulated as follows

.. math::

    \min_{\mathbf{c, y},\epsilon_1, \dots, \epsilon_Q} \quad  & \sum_{q=1}^Q | \epsilon_q |

         \\text{s.t.}\quad\quad  & \mathbf{A'y = c}

         & \mathbf{c'\hat{x}_q = b'y} + \epsilon_q, \quad \\forall q

         & \| \mathbf{c} \|_1 = 1

         & \mathbf{y \geq 0}
"""
import cvxpy as cvx
import numpy as np
#import pudb

from ..utils.invoutils import checkFeasibility, validateFOP


class AbsoluteDualityGap():
    """ Formulate an Absolute Duality Gap method of GMIO.

    Args:
        tol (int): Sets number of significant digits. Default is 8.
        verbose (bool): Sets displays.  Default is False.
        force_feasible_method (bool): If set to True, then will enforce the hyperplane projection method regardless of feasible points. Default is False.
        normalize_c: Set to either 1 or np.inf. Decides the normalization constraint on c
        ban_constraints (list): A list of constraint indices to force to zero when solving. Default is none.

    Example:
        Suppose that the variables ``A`` and ``b`` are numpy matrices and ``points`` is
        a list of numpy arrays::

           model = AbsoluteDualityGap()
           model.FOP(A, b)
           model.solve(points)
           print (model.c)
    """

    def __init__(self, **kwargs):
        self._fop = False
        self._verbose = False
        self._solved = False
        self.tol = 8
        self.solver = cvx.ECOS_BB
        self.force_feasible_method = False
        self.ban_constraints = []
        self.normalize_c = 1
        self._kwargs = self._initialize_kwargs(kwargs)

    def FOP(self, A, b):
        """ Create a forward optimization problem.

        Args:
            A (matrix): numpy matrix of shape :math:`m \\times n`.
            b (matrix): numpy matrix of shape :math:`m \\times 1`.

        Currently, the forward problem is constructed by the user supplying a
        constraint matrix ``A`` and vector ``b``. The forward problem is

        .. math::

            \min_{\mathbf{x}} \quad&\mathbf{c'x}

            \\text{s.t} \quad&\mathbf{A x \geq b}
        """
        #self.A = np.mat(A)
        #self.b = np.mat(b)
        self.A, self.b = validateFOP(A, b)
        self._fop = True

    def solve(self, points, **kwargs):
        """ Solves the inverse optimization problem.

        Args:
            points (list): list of numpy arrays, denoting the (optimal) observed points.

        Returns:
            error (float): the optimal value of the inverse optimization problem.

        First check if all of the points are feasible, in which case we can
        just project the points to each of the hyperplanes. Let :math:`\\bar{x}`
        denote the centroid of the points. Then, we just solve

        .. math::

            \min_{i \in \mathcal{M}} \left\{ \\frac{\mathbf{a_i'\\bar{x} - }b_i }{\| \mathbf{a_i} \|_1} \\right\}

        Let :math:`i^*` denote the optimal index. The optimal cost and dual
        variables are

        .. math::

            \mathbf{c^*} &= \mathbf{\\frac{a_{i^*}}{\|a_{i^*}\|}}

            \mathbf{y^*} &= \mathbf{\\frac{e_{i^*}}{\|a_{i^*}\|}}

        If not all of the points are feasible, then we need to solve an
        exponential number of optimization problems. Let :math:`\mathcal{C}^+, \mathcal{C}^- \subseteq \{ 1, \dots, n \}`
        be a partition of the index set of length ``n``. For each possible
        partition, we solve the following problem

        .. math::

            \min_{\mathbf{c, y}, \epsilon_1,\dots,\epsilon_Q} \quad  & \sum_{q=1}^Q | \epsilon_q |

            \\text{s.t.} \quad  & \mathbf{A'y = c}

            & \mathbf{c'\hat{x}_q = b'y} + \epsilon_q, \quad \\forall q

            & \sum_{i \in \mathcal{C}^+} c_i + \sum_{i \in \mathcal{C}^-} c_i = 1

            & c_i \geq 0, \quad i \in \mathcal{C}^+

            & c_i \leq 0, \quad i \in \mathcal{C}^-

            & \mathbf{y \geq 0}
        """
        self._kwargs = self._initialize_kwargs(kwargs)

        points = [np.mat(point).T for point in points]
        assert self._fop, 'No forward model given.'
        feasible = checkFeasibility(points, self.A, self.b, self.tol)
        if feasible or self.force_feasible_method:
            self.error = self._solveHyperplaneProjection(points)
        else:
            if self.normalize_c == 1:
                self.error = self._solveBruteForceNorm1(points)
            elif self.normalize_c == np.inf:
                self.error = self._solveBruteForceNormInf(points)
            else:
                return -1
        return self.error

    def _solveHyperplaneProjection(self, points):
        m, n = self.A.shape
        errors = np.zeros(m)
        for i in range(m):
            if i in self.ban_constraints:
                errors[i] = 9999999
            else:
                ai = self.A[i] / np.linalg.norm(self.A[i].T, self.normalize_c)
                bi = self.b[i] / np.linalg.norm(self.A[i].T, self.normalize_c)
                errors[i] = np.sum([ai * pt - bi for pt in points])
        minInd = np.argmin(errors)
        self.c = self.A[minInd] / np.linalg.norm(self.A[minInd].T,
                                                 self.normalize_c)
        self.c = self.c.tolist()[0]
        self.error = errors[minInd]
        self.dual = np.zeros(m)
        self.dual[minInd] = 1 / np.linalg.norm(self.A[minInd].T,
                                               self.normalize_c)
        self._solved = True
        return errors[minInd]

    def _baseBruteForceProblem(self, y, z, c, points):
        obj = cvx.Minimize(sum(z))
        cons = []
        cons.append(y >= 0)
        cons.append(self.A.T * y == c)
        for i in range(len(points)):
            chi = self.A * points[i] - self.b
            cons.append(z[i] >= y.T * chi)
            cons.append(z[i] >= -1 * y.T * chi)
        for i in self.ban_constraints:
            cons.append(y[i] == 0)
        return obj, cons

    def _solveBruteForceNorm1(self, points):
        m, n = self.A.shape
        nPoints = len(points)
        nFormulations = 2**n
        bestResult = np.inf

        for formulation in range(nFormulations):
            binFormulation = format(formulation, '0{}b'.format(n))
            cSign = [int(i) for i in binFormulation]
            cSign = np.mat(cSign)
            cSign[cSign == 0] = -1

            y = cvx.Variable(m)
            z = cvx.Variable(nPoints)
            c = cvx.Variable(n)

            obj, cons = self._baseBruteForceProblem(y, z, c, points)

            # add the normalization constraint
            cons.append(cSign * c == 1)
            for i in range(n):
                if cSign[0, i] == 1:
                    cons.append(c[i] >= 0)
                else:
                    cons.append(c[i] <= 0)

            prob = cvx.Problem(obj, cons)
            result = prob.solve(solver=self.solver)

            if result < bestResult:
                bestResult = result
                self.c = c.value / np.linalg.norm(c.value, 1)
                self.dual = y.value / np.linalg.norm(c.value, 1)
        self._solved = True
        self.error = bestResult
        self.dual = self.dual.T.tolist()[0]  # reconvert to just a list
        self.c = self.c.T.tolist()[0]
        return self.error

    def _solveBruteForceNormInf(self, points):
        m, n = self.A.shape
        nPoints = len(points)
        bestResult = np.inf

        for j in range(n):
            y1 = cvx.Variable(m)
            z1 = cvx.Variable(nPoints)
            c1 = cvx.Variable(n)

            obj1, cons1 = self._baseBruteForceProblem(y1, z1, c1, points)

            # Add the normalization constraint
            cons1.append(c1 <= 1)
            cons1.append(c1 >= -1)
            cons1.append(c1[j] == 1)

            prob1 = cvx.Problem(obj1, cons1)
            result1 = prob1.solve(solver=self.solver)

            y2 = cvx.Variable(m)
            z2 = cvx.Variable(nPoints)
            c2 = cvx.Variable(n)

            obj2, cons2 = self._baseBruteForceProblem(y2, z2, c2, points)

            # Add the normalization constraint
            cons2.append(c2 <= 1)
            cons2.append(c2 >= -1)
            cons2.append(c2[j] == -1)

            prob2 = cvx.Problem(obj2, cons2)
            result2 = prob2.solve(solver=self.solver)

            optimalReform = np.argmin([result1, result2, bestResult])
            if optimalReform == 0:
                bestResult = result1
                self.c = c1.value / np.linalg.norm(c1.value, np.inf)
                self.dual = y1.value / np.linalg.norm(y1.value, np.inf)
            elif optimalReform == 1:
                bestResult = result2
                self.c = c2.value / np.linalg.norm(c2.value, np.inf)
                self.dual = y2.value / np.linalg.norm(y2.value, np.inf)

        self._solved = True
        self.error = bestResult
        self.dual = self.dual.T.tolist()[0]  # reconvert to just a list
        self.c = self.c.T.tolist()[0]
        return self.error

    def rho(self, points):
        """ Solves the goodness of fit.
        """
        assert self._solved, 'you need to solve first.'

        m, n = self.A.shape
        numer = [
            np.abs(np.dot(self.c, point) - np.dot(self.dual, self.b))
            for point in points
        ]
        numer = sum(numer)
        denom = 0
        for i in range(m):
            denomTerm = [
                np.abs(np.dot(self.A[i], point) - self.b[i]) / np.linalg.norm(
                    self.A[i].T, self.normalize_c) for point in points
            ]
            denom += sum(denomTerm)
        rho = 1 - numer / denom
        return rho[0, 0]

    def _initialize_kwargs(self, kwargs):
        if 'verbose' in kwargs:
            assert isinstance(kwargs['verbose'],
                              bool), 'verbose needs to be True or False.'
            self._verbose = kwargs['verbose']

        if 'tol' in kwargs:
            assert isinstance(kwargs['tol'],
                              int), 'tolerance needs to be an integer.'
            self.tol = kwargs['tol']

        if 'force_feasible_method' in kwargs:
            assert isinstance(
                kwargs['force_feasible_method'],
                bool), 'force feasible method needs to be True or False.'
            self.force_feasible_method = kwargs['force_feasible_method']

        if 'ban_constraints' in kwargs:
            assert isinstance(kwargs['ban_constraints'],
                              list), 'ban constraints needs to be a list.'
            self.ban_constraints = kwargs['ban_constraints']

        if 'normalize_c' in kwargs:
            assert kwargs['normalize_c'] == 1 or kwargs['normalize_c'] == np.inf, 'normalize c with 1 or infinity norm.'
            self.normalize_c = kwargs['normalize_c']

        if 'solver' in kwargs:
            if kwargs['solver'] in cvx.installed_solvers():
                self.solver = getattr(cvx, kwargs['solver'])
            else:
                print('you do not have this solver.')

        return kwargs



import numpy as np
from scipy.spatial import ConvexHull



def fit_convex_hull(points):
    """ Creates a feasible set by taking a convex hull of the points given. Returns P = { x : Ax >= b }

    Args:
        points (list): Set of numpy points.

    Returns:
        A (numpy): constraint matrix
        b (numpy): constraint vector
    """
    hull = ConvexHull(points)
    m,n = hull.equations.shape
    A = -1 * hull.equations[:,0:n-1]
    b = hull.equations[:,n-1]
    return np.mat(A), np.mat(b).T





""" Relative Duality Gap Inverse Optimization

The relative duality gap uses the ratio of the primal objective value over the
dual objective value of the forward problem  as the measure of sub-optimality.
This inverse optimization problem is formulated as

.. math::

    \min_{\mathbf{c, y},\epsilon_1, \dots, \epsilon_Q} \quad  & \sum_{q=1}^Q | \epsilon_q - 1 |

    \\text{s.t.}\quad\quad  & \mathbf{A'y = c}

    & \mathbf{c'\hat{x}_q = b'y} \epsilon_q

    & \| \mathbf{c} \|_1 = 1

    & \mathbf{y \geq 0}
"""
import cvxpy as cvx
import numpy as np
import pudb

from ..utils.invoutils import checkFeasibility, validateFOP


class RelativeDualityGap():
    """ Formulate an Absolute Duality Gap method of generalized linear inverse
    optimization.

    Args:
        tol (int): Sets number of significant digits. Default is 8.
        verbose (bool): Sets displays.  Default False.
        normalize_c: Set to either 1 or np.inf. Decides the normalization constraint on c
        ban_constraints (list): A list of constraint indices to force to zero when solving. Default is none.

    Example:
        Suppose that the variables ``A`` and ``b`` are numpy matrices and ``points`` is
        a list of numpy arrays::

           model = RelativeDualityGap()
           model.FOP(A, b)
           model.solve(points)
           print (model.c)
    """

    def __init__(self, **kwargs):
        """
        """
        # initialize kwargs
        self._fop = False
        self._solved = False
        self._verbose = False
        self.tol = 8
        self.solver = cvx.ECOS_BB
        self.ban_constraints = []
        self.normalize_c = 1
        self._kwargs = self._initialize_kwargs(kwargs)

    def FOP(self, A, b):
        """ Create a forward optimization problem.

        Args:
            A (matrix): numpy matrix of shape :math:`m \\times n`.
            b (matrix): numpy matrix of shape :math:`m \\times 1`.

        Currently, the forward problem is constructed by the user supplying a
        constraint matrix ``A`` and vector ``b``. The forward problem is

        .. math::

            \min_{\mathbf{x}} \quad&\mathbf{c'x}

            \\text{s.t} \quad&\mathbf{A x \geq b}
        """
        #self.A = np.mat(A)
        #self.b = np.mat(b)
        self.A, self.b = validateFOP(A, b)
        self._fop = True

    def solve(self, points, **kwargs):
        """ Solves the inverse optimization problem.

        Args:
            points (list): list of numpy arrays, denoting the (optimal) observed points.

        Returns:
            error (float): the optimal value of the inverse optimization problem.

        To solve a relative duality gap problem, we solve the three following
        optimization problems.

        .. math::

            \min_{\mathbf{c, y},\epsilon_1, \dots, \epsilon_Q} \quad  & \sum_{q=1}^Q | \epsilon_q - 1 |

            \\text{s.t.}\quad\quad  & \mathbf{A'y = c}

            & \mathbf{c'\hat{x}_q = } \epsilon_q

            & \mathbf{b'y} = 1

            & \mathbf{y \geq 0}

        .. math::

            \min_{\mathbf{c, y},\epsilon_1, \dots, \epsilon_Q} \quad  & \sum_{q=1}^Q | \epsilon_q - 1 |

            \\text{s.t.}\quad\quad  & \mathbf{A'y = c}

            & \mathbf{c'\hat{x}_q = } -\epsilon_q

            & \mathbf{b'y} = -1

            & \mathbf{y \geq 0}

        .. math::

            \min_{\mathbf{c, y},\epsilon_1, \dots, \epsilon_Q} \quad  & 0

            \\text{s.t.}\quad\quad  & \mathbf{A'y = c}

            & \mathbf{c'\hat{x}_q =} 0

            & \mathbf{b'y} = 0

            & \mathbf{y'1} = 0

            & \mathbf{y \geq 0}

        The optimal value of the relative duality gap problem is equal to the
        optimal value of the minimum of these problems. Let :math:`\mathbf{\hat{c}, \hat{y}}`
        denote the optimal solution of that corresponding problem. Then, the
        optimal solution of the relative duality gap problem is

        .. math::

            \mathbf{c^*} &= \mathbf{\\frac{\hat{c}}{\|\hat{c}\|_1}}

            \mathbf{y^*} &= \mathbf{\\frac{\hat{y}}{\|\hat{c}\|_1}}
        """
        self._kwargs = self._initialize_kwargs(kwargs)

        points = [np.mat(point).T for point in points]
        assert self._fop, 'No forward model given.'
        self.error = self._solveRelativeDGLP(points)
        if (np.round(self.c, self.tol) == 0).all():
            # the relative problem returned c = 0 and we have to fix.
            # First we solve the auxiliary problem
            self.K = self._solve_auxiliary_problem()

            #Then we solve the "hard way"
            self.error = self._solveRelativeDG(points)
        return self.error

    def _solve_auxiliary_problem(self):
        """ Solves the auxiliary problem.

            max max(b'y, -b'y, y'1)
            st  ||A'y|| = 1
                y >= 0
        """
        if self.normalize_c == 1:
            res = self._solve_auxiliary_problemNorm1()
        elif self.normalize_c == np.inf:
            res = self._solve_auxiliary_problemNormInf()
        return 1 / res

    def _solve_auxiliary_problemNorm1(self):
        """ Solves the auxiliary problem with || A'y ||_1 = 1
        """
        m, n = self.A.shape
        nFormulations = 2**n
        bestResult = 0

        for formulation in range(nFormulations):
            binFormulation = format(formulation, '0{}b'.format(n))
            cSign = [int(i) for i in binFormulation]
            cSign = np.mat(cSign)
            cSign[cSign == 0] = -1

            y1 = cvx.Variable(m)
            c1 = cvx.Variable(n)
            cons1 = []
            cons1.append(self.A.T * y1 == c1)
            cons1.append(y1 >= 0)
            cons1.append(cSign * c1 == 1)

            for i in range(n):
                if cSign[0, i] == 1:
                    cons.append(c1[i] >= 0)
                else:
                    cons.append(c1[i] <= 0)

            obj1 = cvx.Maximize(self.b.T * y1)
            prob1 = cvx.Problem(obj1, cons2)
            result1 = prob1.solve(solver=self.solver)

            y2 = cvx.Variable(m)
            c2 = cvx.Variable(n)
            cons2 = []
            cons2.append(self.A.T * y2 == c2)
            cons2.append(y2 >= 0)
            cons1.append(cSign * c2 == 1)

            for i in range(n):
                if cSign[0, i] == 1:
                    cons.append(c2[i] >= 0)
                else:
                    cons.append(c2[i] <= 0)

            obj2 = cvx.Maximize(-1 * self.b.T * y2)
            prob2 = cvx.Problem(obj2, cons2)
            result2 = prob2.solve(solver=self.solver)

            y3 = cvx.Variable(m)
            c3 = cvx.Variable(n)
            cons3 = []
            cons3.append(self.A.T * y3 == c3)
            cons3.append(y3 >= 0)
            cons1.append(cSign * c3 == 1)

            for i in range(n):
                if cSign[0, i] == 1:
                    cons.append(c3[i] >= 0)
                else:
                    cons.append(c3[i] <= 0)

            allOnes = np.mat(np.ones(m)).T
            obj3 = cvx.Maximize(y3.T * allOnes)
            prob3 = cvx.Problem(obj3, cons3)
            result3 = prob3.solve(solver=self.solver)

            result = np.max([result1, result2, result3])

            if result > bestResult:
                bestResult = result
        return result

    def _solve_auxiliary_problemNormInf(self):
        """ Solves the auxiliary problem with || A'y ||_infinty = 1
        """
        m, n = self.A.shape
        bestResult = 0

        for j in range(n):
            y1 = cvx.Variable(m)
            c1 = cvx.Variable(n)
            cons1 = []
            cons1.append(self.A.T * y1 == c1)
            cons1.append(y1 >= 0)

            cons1.append(c1[j] == 1)
            cons1.append(c1 <= 1)
            cons1.append(c1 >= -1)

            obj1 = cvx.Maximize(self.b.T * y1)
            prob1 = cvx.Problem(obj1, cons1)
            result1 = prob1.solve(solver=self.solver)

            y2 = cvx.Variable(m)
            c2 = cvx.Variable(n)
            cons2 = []
            cons2.append(self.A.T * y2 == c2)
            cons2.append(y2 >= 0)

            cons2.append(c2[j] == 1)
            cons2.append(c2 <= 1)
            cons2.append(c2 >= -1)

            obj2 = cvx.Maximize(-1 * self.b.T * y2)
            prob2 = cvx.Problem(obj2, cons2)
            result2 = prob2.solve(solver=self.solver)

            y3 = cvx.Variable(m)
            c3 = cvx.Variable(n)
            cons3 = []
            cons3.append(self.A.T * y3 == c3)
            cons3.append(y3 >= 0)

            cons3.append(c3[j] == 1)
            cons3.append(c3 <= 1)
            cons3.append(c3 >= -1)

            allOnes = np.mat(np.ones(m)).T
            obj3 = cvx.Maximize(y3.T * allOnes)
            prob3 = cvx.Problem(obj3, cons3)
            result3 = prob3.solve(solver=self.solver)

            result = np.max([result1, result2, result3])

            if result > bestResult:
                bestResult = result
        return result

    def _baseBruteForceProblem(self, y, z, c):
        obj = cvx.Minimize(sum(z))
        cons = []
        cons.append(y >= 0)
        cons.append(self.A.T * y == c)
        for i in self.ban_constraints:
            cons.append(y[i] == 0)
        return obj, cons

    def _solveRelativeDG(self, points):
        """ Solves the norm constrained version of the problem.

            min sum z_q
            st  z_q >= c'x_q - 1
                z_q >= 1 - c'x_q
                A'y = c
                b'y = 1
                ||c|| = 1
                y >= 0
        """
        if self.normalize_c == 1:
            error = self._solveRelativeDGNorm1(points)
        elif self.normalize_c == np.inf:
            error = self._solveRelativeDGNormInf(points)
        return error

    def _solveRelativeDGNorm1(self, points):
        """ Solves the problem with the l-1 norm.
        """
        m, n = self.A.shape
        nPoints = len(points)
        nFormulations = 2**n
        bestResult = np.inf

        for formulation in range(nFormulations):
            binFormulation = format(formulation, '0{}b'.format(n))
            cSign = [int(i) for i in binFormulation]
            cSign = np.mat(cSign)
            cSign[cSign == 0] = -1

            # first solve the positive variant
            obj1, cons1, y1, z1, c1 = self._positiveProblem(
                m, n, nPoints, points)
            cons1.append(cSign * c1 >= self.K)
            for i in range(n):
                if cSign[0, i] == 1:
                    cons1.append(c1[i] >= 0)
                else:
                    cons1.append(c1[i] <= 0)
            prob1 = cvx.Problem(obj1, cons1)
            result1 = prob1.solve(solver=self.solver)

            # then solve the negative variant
            obj2, cons2, y2, z2, c2 = self._negativeProblem(
                m, n, nPoints, points)
            cons2.append(cSign * c2 >= self.K)
            for i in range(n):
                if cSign[0, i] == 1:
                    cons2.append(c2[i] >= 0)
                else:
                    cons2.append(c2[i] <= 0)
            prob2 = cvx.Problem(obj2, cons2)
            result2 = prob2.solve(solver=self.solver)

            # then solve the zero variant
            obj3, cons3, y3, c3 = self._zeroProblem(m, n, nPoints, points)
            cons3.append(cSign * c3 >= self.K)
            for i in range(n):
                if cSign[0, i] == 1:
                    cons3.append(c3[i] >= 0)
                else:
                    cons3.append(c3[i] <= 0)
            prob3 = cvx.Problem(obj3, cons3)
            result3 = prob3.solve(solver=self.solver)

            best = np.argmin([result1, result2, result3, bestResult])
            if best == 0:
                bestResult = result1
                self.c = c1.value / np.linalg.norm(c1.value, 1)
                self.dual = y1.value / np.linalg.norm(y1.value, 1)
            elif best == 1:
                bestResult = result2
                self.c = c2.value / np.linalg.norm(c2.value, 1)
                self.dual = y2.value / np.linalg.norm(y2.value, 1)
            elif best == 2:
                bestResult = result3
                self.c = c3.value / np.linalg.norm(c3.value, 1)
                self.dual = y3.value / np.linalg.norm(y3.value, 1)
        self._solved = True
        self.error = bestResult
        self.dual = self.dual.T.tolist()[0]  # reconvert to just a list
        self.c = self.c.T.tolist()[0]
        return self.error

    def _solveRelativeDGNormInf(self, points):
        """ Solves the problem with the l-infinity norm.
        """
        m, n = self.A.shape
        nPoints = len(points)
        bestResult = np.inf

        for j in range(n):
            # first solve the positive variant
            obj1p, cons1p, y1p, z1p, c1p = self._positiveProblem(
                m, n, nPoints, points)
            cons1p.append(c1p[j] >= self.K)
            prob1p = cvx.Problem(obj1p, cons1p)
            result1p = prob1p.solve(solver=self.solver)

            obj1n, cons1n, y1n, z1n, c1n = self._positiveProblem(
                m, n, nPoints, points)
            cons1n.append(c1n[j] <= -self.K)
            prob1n = cvx.Problem(obj1p, cons1p)
            result1n = prob1n.solve(solver=self.solver)

            # then solve the negative variant
            obj2p, cons2p, y2p, z2p, c2p = self._negativeProblem(
                m, n, nPoints, points)
            cons2p.append(c2p[j] >= self.K)
            prob2p = cvx.Problem(obj2p, cons2p)
            result2p = prob2p.solve(solver=self.solver)

            obj2n, cons2n, y2n, z2n, c2n = self._negativeProblem(
                m, n, nPoints, points)
            cons2n.append(c2n[j] <= -self.K)
            prob2n = cvx.Problem(obj1p, cons1p)
            result2n = prob2n.solve(solver=self.solver)

            # then solve the zero variant
            obj3p, cons3p, y3p, c3p = self._zeroProblem(m, n, nPoints, points)
            cons3p.append(c3p[j] >= self.K)
            prob3p = cvx.Problem(obj3p, cons3p)
            result3p = prob3p.solve(solver=self.solver)

            obj3n, cons3n, y3n, c3n = self._zeroProblem(m, n, nPoints, points)
            cons3n.append(c3n[j] <= -self.K)
            prob3n = cvx.Problem(obj1p, cons1p)
            result3n = prob3n.solve(solver=self.solver)

            best = np.argmin([
                result1p, result1n, result2p, result2n, result3p, result3n,
                bestResult
            ])
            if best == 0:
                bestResult = result1p
                self.c = c1p.value / np.linalg.norm(c1p.value, np.inf)
                self.dual = y1p.value / np.linalg.norm(y1p.value, np.inf)
            elif best == 1:
                bestResult = result1n
                self.c = c1n.value / np.linalg.norm(c1n.value, np.inf)
                self.dual = y1n.value / np.linalg.norm(y1n.value, np.inf)
            elif best == 2:
                bestResult = result2p
                self.c = c2p.value / np.linalg.norm(c2p.value, np.inf)
                self.dual = y2p.value / np.linalg.norm(y2p.value, np.inf)
            elif best == 3:
                bestResult = result2n
                self.c = c2n.value / np.linalg.norm(c2n.value, np.inf)
                self.dual = y2n.value / np.linalg.norm(y2n.value, np.inf)
            elif best == 4:
                bestResult = result3p
                self.c = c3p.value / np.linalg.norm(c3p.value, np.inf)
                self.dual = y3p.value / np.linalg.norm(y3p.value, np.inf)
            elif best == 5:
                bestResult = result3n
                self.c = c3n.value / np.linalg.norm(c3n.value, np.inf)
                self.dual = y3n.value / np.linalg.norm(y3n.value, np.inf)
        self._solved = True
        self.error = bestResult
        self.dual = self.dual.T.tolist()[0]  # reconvert to just a list
        self.c = self.c.T.tolist()[0]
        return self.error

    def _positiveProblem(self, m, n, nPoints, points):
        y = cvx.Variable(m)
        z = cvx.Variable(nPoints)
        c = cvx.Variable(n)
        obj, cons = self._baseBruteForceProblem(y, z, c)
        cons.append(y.T * self.b == 1)
        for i, point in enumerate(points):
            cons.append(z[i] >= c.T * point - 1)
            cons.append(z[i] >= 1 - c.T * point)
        return obj, cons, y, z, c

    def _negativeProblem(self, m, n, nPoints, points):
        y = cvx.Variable(m)
        z = cvx.Variable(nPoints)
        c = cvx.Variable(n)
        obj, cons = self._baseBruteForceProblem(y, z, c)
        cons.append(y.T * self.b == -1)
        for i, point in enumerate(points):
            cons.append(z[i] >= c.T * point + 1)
            cons.append(z[i] >= -1 - c.T * point)
        return obj, cons, y, z, c

    def _zeroProblem(self, m, n, nPoints, points):
        y = cvx.Variable(m)
        z = cvx.Variable(nPoints)
        c = cvx.Variable(n)
        _, cons = self._baseBruteForceProblem(y, z, c)
        obj = cvx.Minimize(0)
        cons.append(y.T * self.b == 0)
        allOnes = np.mat(np.ones(m)).T
        cons.append(y.T * allOnes == 1)
        for point in points:
            cons.append(c.T * point == 0)
        return obj, cons, y, c

    def _solveRelativeDGLP(self, points):
        """ Solves a linear program.

            min sum z_q
            st  [ z_q >= c'x_q - 1
            z_q >= 1 - c'x_q
            A'y = c
            b'y = 1
            y >= 0 ]
        """
        m, n = self.A.shape
        nPoints = len(points)

        # first solve the positive variant
        obj1, cons1, y1, z1, c1 = self._positiveProblem(m, n, nPoints, points)
        prob1 = cvx.Problem(obj1, cons1)
        result1 = prob1.solve(solver=self.solver)

        # then solve the negative variant
        obj2, cons2, y2, z2, c2 = self._negativeProblem(m, n, nPoints, points)
        prob2 = cvx.Problem(obj2, cons2)
        result2 = prob2.solve(solver=self.solver)

        # then solve the zero variant
        obj3, cons3, y3, c3 = self._zeroProblem(m, n, nPoints, points)
        prob3 = cvx.Problem(obj3, cons3)
        result3 = prob3.solve(solver=self.solver)

        optimalReform = np.argmin([result1, result2, result3])
        if optimalReform == 0:
            self.c = c1.value / np.linalg.norm(c1.value, self.normalize_c)
            self.dual = y1.value / np.linalg.norm(y1.value, self.normalize_c)
        elif optimalReform == 1:
            self.c = c2.value / np.linalg.norm(c2.value, self.normalize_c)
            self.dual = y2.value / np.linalg.norm(y2.value, self.normalize_c)
        elif optimalReform == 2:
            self.c = c3.value / np.linalg.norm(c3.value, self.normalize_c)
            self.dual = y3.value / np.linalg.norm(y3.value, self.normalize_c)

        self.c = self.c.T.tolist()[0]
        self.dual = self.dual.T.tolist()[0]
        self._solved = True
        return np.min([result1, result2, result3])

    def rho(self, points):
        """ Solves the goodness of fit.
        """
        assert self._solved, 'you need to solve first.'

        m, n = self.A.shape
        #numer = [ np.abs(np.dot(self.c, point) - np.dot(self.dual, self.b)) / np.abs(np.dot(self.dual, self.b)) for point in points ]
        numer = [np.abs(np.dot(self.c, point) - 1) for point in points]
        numer = sum(numer)
        denom = 0
        for i in range(m):
            #denomTerm = [ np.abs(np.dot(self.A[i], point) - self.b[i]) / np.abs(self.b[i]) for point in points ]
            denomTerm = [
                np.abs(
                    np.dot(self.A[i] / np.linalg.norm(
                        self.A[i].T, self.normalize_c), point) - 1)
                for point in points
            ]
            denom += sum(denomTerm)
        rho = 1 - numer / denom
        return rho[0, 0]

    def _initialize_kwargs(self, kwargs):
        if 'verbose' in kwargs:
            assert isinstance(kwargs['verbose'],
                              bool), 'verbose needs to be True or False.'
            self._verbose = kwargs['verbose']

        if 'tol' in kwargs:
            assert isinstance(kwargs['tol'],
                              int), 'tolernace needs to be an integer.'
            self.tol = kwargs['tol']

        if 'ban_constraints' in kwargs:
            assert isinstance(kwargs['ban_constraints'],
                              list), 'ban constraints needs to be a list.'
            self.ban_constraints = kwargs['ban_constraints']

        if 'normalize_c' in kwargs:
            assert kwargs['normalize_c'] == 1 or kwargs['normalize_c'] == np.inf, 'normalize c with 1 or infinity norm.'
            self.normalize_c = kwargs['normalize_c']

        if 'solver' in kwargs:
            if kwargs['solver'] in cvx.installed_solvers():
                self.solver = getattr(cvx, kwargs['solver'])
            else:
                print('you do not have this solver.')

        return kwargs



import numpy as np




def checkFeasibility(points, A, b, tol=8):
    """ Check if the set of points satisfy A x >= b.
    
    Args:
        points (list): set of numpy points.
        A (list): numpy matrix.
        b (list): numpy vector.
    """
    def _checkFeasibilityPoint(point, A, b, tol):
        check = np.round(A * point - b, tol)
        return (check >= 0).all()
    
    allPointsAreFeasible = np.array([ _checkFeasibilityPoint(point, A, b, tol) for point in points]).all()
    return allPointsAreFeasible


def validateFOP(A, b):
    """ Validate that A, b are numpy matrices with the appropriate structure.
    """
    aType = type(A)
    bType = type(b)
    if aType is np.matrixlib.defmatrix.matrix:
        m, n = A.shape
        Aout = A

    elif aType is np.ndarray:
        m, n = A.shape
        Aout = np.matrix(A)
    
    elif aType is list:
        m = len(A)
        n = len(A[0])
        for ai in A:
            if len(ai) != n:
                raise ValueError('Rows of A are incomplete')
        aRowType = type(A[0])
        
        if aRowType is np.matrixlib.defmatrix.matrix:
            Aout = []
            for ai in A:
                Aout.append(ai.tolist())[0]
        elif aRowType is np.ndarray:
            Aout = []
            for ai in A:
                Aout.append(ai.tolist())[0]
        elif aRowType is list:
            Aout = A
        
        else:
            raise TypeError('Type of A is incorrect.')
        Aout = np.matrix(Aout)

    else:
        raise TypeError('Type of A is incorrect.')

    if bType is np.matrixlib.defmatrix.matrix:
        m2, n2 = b.shape
        if m2 != m:
            bout = b.T
        else:
            bout = b

    elif bType is np.ndarray:
        if len(b.shape) == 1:
            bout = np.mat(b).T
        else:
            raise TypeError('Type of b is incorrect.')

    elif bType is list:
            bout = np.mat(b).T

    else:
        raise TypeError('Type of b is incorrect.')


    # do a final validation here.
    m, n = Aout.shape
    m2, n2 = bout.shape
    if m != m2:
        raise TypeError('Type of A or b is incorrect.')
    if n2 != 1:
        raise TypeError('Type of b is incorrect.')
    return Aout, bout



def validatePoints(points):
    pType = type(points)

    if pType is list:
        p2Type = type(points[0])
        pOut = []
        for point in points:
            if type(point) != p2Type:
                raise TypeError('Points are not all the same type.')
            if p2Type is list:
                pOut.append(np.mat(point).T)
            elif p2Type is np.matrixlib.defmatrix.matrix:
                if point.shape[0] == 1: 
                    pOut.append(point.T)
                else:
                    pOut.append(point)
            elif p2Type is np.ndarray:
                if len(p3.shape) == 1:
                    pOut.append(np.mat(point).T)
                else:
                    raise TypeError('Type of points is incorrect.')
            else:
                raise TypeError('Type of points is incorrect.')

    elif pType is np.matrixlib.defmatrix.matrix:
        pass

    elif pType is np.ndarray:
        pass

    else:
        raise TypeError('Type of points is incorrect.')



""" invo - An Inverse Optimization Library

invo is a Python package intended to simplify the use of inverse optimization as
a model fitting practice. Our goal is to provide a general framework for 
formulating and solving inverse optimization problems, as well as implement a
collection of inverse methods. As additional algorithms and methods arise in 
the literature, we can implement them using the same common framework.

The flavour of this work is inspired from scikit-learn on Python.
"""



__all__ = ['LinearModels', 'utils']




#! usr/bin/env/python
#
# By Rafid Mahmood <rafid.mahmood@mail.utoronto.ca>
# License: MIT

from setuptools import setup

setup(
    name='invo',
    version='0.5',
    description='A Generalized Inverse Optimization package.',
    author='Rafid Mahmood',
    author_email='rafid.mahmood@mail.utoronto.ca',
    url='https://github.com/rafidrm/invo',
    download_url='https://github.com/rafidrm/invo/archive/0.5.tar.gz',
    license='MIT',
    packages=['invo', 'invo.LinearModels', 'invo.utils'],
    keywords=['optimization', 'inverse'],
    install_requires=['numpy', 'cvxpy'],
    zip_safe=False)
